# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from avenue import Avenue


class TestAvenue(object):
    def test_basics(self):
        router = Avenue()

        @router.attach(path='/', method='GET')
        @router.attach(path='/welcome')
        def route_1():
            return 'Route 1'

        @router.attach(path='/', method='POST')
        def route_2():
            return 'Route 2'

        @router.attach(path='/post/<id|int>')
        def route_3(id):
            return 'Route 3 - for {:s}'.format(repr(id))

        @router.attach(path='/post/new')
        def route_4():
            return 'Route 4'

        @router.attach(path='/<page|greedy>')
        def route_5(page):
            return 'Route 5 - fallback - {:s}'.format(page)

        routing = [
            [{'path': '/', 'method': 'GET'},
             'Route 1'],
            [{'path': '/', 'method': 'POST'},
             'Route 2'],
            [{'path': '/post/12', 'method': 'GET'},
             'Route 3 - for 12'],
            [{'path': '/post/new-article-here', 'method': 'GET'},
             'Route 5 - fallback - post/new-article-here'],
            [{'path': '/post/new', 'method': 'GET'},
             'Route 4'],
            [{'path': '/any/other/page', 'method': 'GET'},
             'Route 5 - fallback - any/other/page'],
        ]

        for route, result in routing:
            path, match = router.match(**route)
            assert path.func(**match.kwargs) == result



from __future__ import absolute_import, division, print_function

from pandora import CONVERTERS
from pandora import Path


class Match(object):
    def __init__(self, score=None, kwargs=None):
        self.score = score or list()
        self.kwargs = kwargs or dict()

    def update(self, result):
        if type(result) is int:
            self.score.append(result)
        elif isinstance(result, Match):
            self.score.append(result.score)
            self.kwargs.update(result.kwargs)

    def __lt__(self, other):
        if isinstance(other, Match):
            return self.score < other.score
        else:
            return self.score < other

    def __repr__(self):
        return '<Match (score={:s})>'.format(repr(self.score))


class MatchError(Exception):
    pass


class Processor(object):
    def __init__(self, name, optional=False):
        self.name = name
        self.optional = optional

    @classmethod
    def argument(cls, route):
        return route.kwargs[cls.__name__.replace('Processor', '').lower()]


class PathProcessor(Processor):
    def __init__(self, name, optional=False, converters=None):
        super(PathProcessor, self).__init__(name, optional)
        self.cache = dict()
        self.converters = converters or CONVERTERS

    def __call__(self, value, route):
        """

        :type value: mixed
        :type route: avenue.Route
        :rtype: Match
        """
        try:
            argument = self.argument(route)
        except KeyError:
            if self.optional:
                return Match(0)
        else:
            if argument not in self.cache:
                self.cache[argument] = Path(argument, self.converters,
                                            split='/')

            path = self.cache[argument]
            match = path(value)

            if match is not None:
                return Match(path, match)

            raise MatchError('Path does not match, `{:s}` != `{:s}`'.format(
                    path.original, argument))


class MethodProcessor(Processor):
    def __call__(self, value, route):
        """

        :type value: mixed
        :type route: avenue.Route
        :rtype: Match
        """
        try:
            argument = self.argument(route)
        except KeyError:
            if self.optional:
                return 0
        else:
            if value == argument:
                return 1
            else:
                error = 'Method does not match, `{:s}` != `{:s}`'.format(
                        value, argument)
                raise MatchError(error)



from __future__ import absolute_import, division, print_function

from .processors import Match
from .processors import MatchError
from .processors import MethodProcessor
from .processors import PathProcessor
import operator

PROCESSORS = [
    PathProcessor('path'),
    MethodProcessor('method', optional=True),
]


class Avenue(object):
    def __init__(self, processors=None):
        self.routes = list()
        self.processors = processors or PROCESSORS

    def attach(self, **kwargs):
        def inner(func):
            self.routes.append(Route(self, kwargs, func))
            return func

        return inner

    def match(self, **kwargs):
        matches = list()

        for route in self.routes:
            try:
                match = Match()

                for processor in self.processors:
                    value = kwargs.get(processor.name)
                    match.update(processor(value, route))
            except MatchError:
                pass
            else:
                if len(match.score):
                    matches.append([route, match])

        size = max(map(len, [m[1].score for m in matches]))
        matches = list(filter(lambda m: len(m[1].score) == size, matches))
        matches.sort(key=operator.itemgetter(1))

        if len(matches):
            return matches[0][0], matches[0][1]
        else:
            return None, None


class Route(object):
    def __init__(self, avenue, kwargs, func):
        self.avenue = avenue
        self.kwargs = kwargs
        self.func = func
